@@ -362,6 +362,44 @@ ReplicatorSMM::donate_sst(void* const         recv_ctx,
     return ret;
 }
 
+struct slg
+{
+    gcache::GCache& gcache_;
+    bool            unlock_;
+
+    slg(gcache::GCache& cache) : gcache_(cache), unlock_(false){}
+    ~slg() { if (unlock_) gcache_.seqno_unlock(); }
+};
+
+static wsrep_seqno_t run_ist_senders(ist::AsyncSenderMap& ist_senders,
+                                     const gu::Config&    config,
+                                     const std::string&   peer,
+                                     wsrep_seqno_t const  preload_start,
+                                     wsrep_seqno_t const  cc_seqno,
+                                     wsrep_seqno_t const  cc_lowest,
+                                     int const            proto_ver,
+                                     slg&                 seqno_lock_guard,
+                                     wsrep_seqno_t const  rcode)
+{
+    try
+    {
+        ist_senders.run(config,
+                        peer,
+                        preload_start,
+                        cc_seqno,
+                        cc_lowest,
+                        proto_ver);
+        // seqno will be unlocked when sender exists
+        seqno_lock_guard.unlock_ = false;
+        return rcode;
+    }
+    catch (gu::Exception& e)
+    {
+        log_warn << "IST failed: " << e.what();
+        return -e.get_errno();
+    }
+}
+
 void ReplicatorSMM::process_state_req(void*       recv_ctx,
                                       const void* req,
                                       size_t      req_size,
@@ -408,15 +446,7 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
 
     if (not skip_sst)
     {
-        struct sgl
-        {
-            gcache::GCache& gcache_;
-            bool            unlock_;
-
-            sgl(gcache::GCache& cache) : gcache_(cache), unlock_(false){}
-            ~sgl() { if (unlock_) gcache_.seqno_unlock(); }
-        }
-        seqno_lock_guard(gcache_);
+        slg seqno_lock_guard(gcache_);
 
         if (streq->ist_len())
         {
@@ -450,7 +480,7 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
                     wsrep_gtid_t const state_id =
                         { istr.uuid(), istr.last_applied() };
 
-                    rcode = donate_sst(recv_ctx, *streq, state_id, true);
+                    gu_trace(rcode = donate_sst(recv_ctx, *streq, state_id, true));
 
                     // we will join in sst_sent.
                     join_now = false;
@@ -458,26 +488,19 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
 
                 if (rcode >= 0)
                 {
-                    try
-                    {
-                        ist_senders_.run(config_,
-                                         istr.peer(),
-                                         first,
-                                         cc_seqno_,
-                                         cc_lowest_trx_seqno_,
-                                         /* Historically IST messages versioned
-                                          * with the global replicator protocol.
-                                          * Need to keep it that way for backward
-                                          * compatibility */
-                                         protocol_version_);
-                        // seqno will be unlocked when sender exists
-                        seqno_lock_guard.unlock_ = false;
-                    }
-                    catch (gu::Exception& e)
-                    {
-                        log_error << "IST failed: " << e.what();
-                        rcode = -e.get_errno();
-                    }
+                    rcode = run_ist_senders(ist_senders_,
+                                            config_,
+                                            istr.peer(),
+                                            first,
+                                            cc_seqno_,
+                                            cc_lowest_trx_seqno_,
+                        /* Historically IST messages are versioned
+                         * with the global replicator protocol.
+                         * Need to keep it that way for backward
+                         * compatibility */
+                                            protocol_version_,
+                                            seqno_lock_guard,
+                                            rcode);
                 }
                 else
                 {
@@ -553,18 +576,20 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
                     IST_request istr;
                     get_ist_request(streq, &istr);
                     // Send trxs to rebuild cert index.
-                    ist_senders_.run(config_,
-                                     istr.peer(),
-                                     preload_start,
-                                     cc_seqno_,
-                                     preload_start,
-                                     /* Historically IST messages are versioned
-                                      * with the global replicator protocol.
-                                      * Need to keep it that way for backward
-                                      * compatibility */
-                                     protocol_version_);
-                    // seqno will be unlocked when sender exists
-                    seqno_lock_guard.unlock_ = false;
+                    rcode = run_ist_senders(ist_senders_,
+                                            config_,
+                                            istr.peer(),
+                                            preload_start,
+                                            cc_seqno_,
+                                            preload_start,
+                        /* Historically IST messages are versioned
+                         * with the global replicator protocol.
+                         * Need to keep it that way for backward
+                         * compatibility */
+                                            protocol_version_,
+                                            seqno_lock_guard,
+                                            rcode);
+                    if (rcode < 0) goto out;
                 }
                 else /* streq->version() == 0 */
                 {
