@@ -14,6 +14,8 @@
 #include "gu_uri.hpp"
 #include "gu_signals.hpp"
 
+#include "wsrep_allowlist_service.h"
+
 #include <netinet/tcp.h> // tcp_info
 
 #include <array>
@@ -758,6 +760,13 @@ namespace gu
         class Impl;
         std::unique_ptr<Impl> impl_;
     };
+
+    /* Allowlist check callback */
+    bool allowlist_value_check(wsrep_allowlist_key_t key, const std::string& value);
+
+    /* Init/deinit global allowlist service hooks. */
+    int init_allowlist_service_v1(wsrep_allowlist_service_v1_t*);
+    void deinit_allowlist_service_v1();
 }
 
 #endif // GU_ASIO_HPP
