@@ -1,4 +1,4 @@
-/* Copyright (C) 2011-2016 Codership Oy <info@codership.com> */
+/* Copyright (C) 2011-2020 Codership Oy <info@codership.com> */
 
 #include "garb_recv_loop.hpp"
 
@@ -117,6 +117,10 @@ RecvLoop::loop()
 
             break;
         }
+        case GCS_ACT_INCONSISTENCY:
+            // something went terribly wrong, restart needed
+            gcs_.close();
+            return;
         case GCS_ACT_JOIN:
         case GCS_ACT_SYNC:
         case GCS_ACT_FLOW:
