@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2015 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2021 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -732,7 +732,7 @@ int main (int argc, char *argv[])
     gcs_register_params(gconf);
 
     if (!(cache = gcache_create (gconf, ""))) goto out;
-    if (!(gcs = gcs_create (gconf, cache, NULL, NULL, 0, 0))) goto out;
+    if (!(gcs = gcs_create (gconf, cache, NULL, NULL, NULL, 0, 0))) goto out;
     puts ("debug"); fflush(stdout);
     /* the following hack won't work if there is 0.0.0.0 in URL options */
     bstrap = (NULL != strstr(conf.backend, "0.0.0.0"));
